#
# aviwriter.tcl
#
# A sample script to create AVI file by giving a sequence of images.
#
# Warning:
# * There is no guarranty. 
#
# * This script runs very slowly. This is not practical.
#
#   Instead of the inefficient pixel-by-pixel traversal of nested loops,
#   binary data dump by "image data -format ppm" is used for efficiency.
#   Yet, remaining extra processings such as
#     top-bottom flip, base64-decoding, r-g-b flip,  0 insertion...
#   still slow down the operation badly.
#
# * I suggest saving images to files and using a converter application
#   to generate a movie out of them.
#
# Usage:
#     source "aviwriter.tcl"
#     set aw [AVIWriter \#auto]
#     $aw fopen "output.avi"
#     $aw putFrame image1  <= the first image determine the width and the height
#     $aw putFrame image2
#     :
#     $aw setFPS 30        <= FPS can be set at any locations before fclose.
#     $aw fclose
#
# Rev. 0.1. Feb. 12, 2011  zhuo
#

package require Itcl
package require Tk
package require Img
package require base64

# avi header section snapshot
# "R I F F <=size> A V I   L I S T"
# "sz=c0h  h d r l a v i h sz=38h "
# "IFRAME  MAXDR   align   flags  ", iframe = interval of frame, in usec. 1000000/FPS. here,33333.
# "NFRAME  begfr   nstrm   BUFSZ"      *maxdr : W x H x 4 x FPS.
# "WIDTH   HEIGHT  resv    resv"       *NFRAME : total num frame
# "resv    resv    L I S T sz=74h "    *BUFSZ : W x H x 4
# "s t r l s t r h sz=38h  v i d s"
# "fcch    Flag    lang    intfram" ,   fcch = "DIB "
# "SCALE   rate    start   NFRAME ",    SCALE=IFRAME=0x8235=33333, RATE=40420f=1000000; 
# "BUFSZ   quality samplsz (pad)  "
# "(pad)   s t r f sz=28h  sz=28h "
# "WIDTH   HEIGHT  pln/bitcnt compr"
# "SZIMAGE XPels/M YPels/M  ClrUsd"
# "ClrImpt L I S T <vsize> m o v i"
#
# one image:
# "0 0 d b SZIMAGE (b)(g)(r)0....."
#
# index:
# "i d x 1 <vsize> 0 0 d b sz=10h "
# "loc.    SZIMAGE 0 0 d b sz=10h "
# "loc.    SZIMAGE ....

#A1 : size of RIFF
#A2 : Interval of Frame, in usec
#A3 : number of frames
#A4 : BUFSZ=SZIMAGE 
#A5 : Width.
#A6 : Height:
#A7 : size of LIST for movi's 
#A8 : max data rate

itcl::class AVIWriter {
	private common headerTemplateBin_
	set headerTemplateBin_ [binary format H* [join [split "
	52 49 46 46 A1 A1 A1 A1 41 56 49 20 4c 49 53 54
	c0 00 00 00 68 64 72 6c 61 76 69 68 38 00 00 00
	A2 A2 A2 A2 A8 A8 A8 A8 00 00 00 00 10 08 00 00
	A3 A3 A3 A3 00 00 00 00 01 00 00 00 A4 A4 A4 A4
	A5 A5 A5 A5 A6 A6 A6 A6 00 00 00 00 00 00 00 00
    00 00 00 00 00 00 00 00 4c 49 53 54 74 00 00 00
    73 74 72 6c 73 74 72 68 38 00 00 00 76 69 64 73
    44 49 42 20 00 00 00 00 00 00 00 00 00 00 00 00
    A2 A2 A2 A2 40 42 0f 00 00 00 00 00 A3 A3 A3 A3
    A4 A4 A4 A4 ff ff ff ff 00 00 00 00 00 00 00 00
    A5 A5 A6 A6 73 74 72 66 28 00 00 00 28 00 00 00
    A5 A5 A5 A5 A6 A6 A6 A6 01 00 20 00 00 00 00 00
    A4 A4 A4 A4 00 00 00 00 00 00 00 00 00 00 00 00
    00 00 00 00 4c 49 53 54 A7 A7 A7 A7 6d 6f 76 69" ] "" ] ]
	private variable fps_

	private variable fd_
	private variable fname_

	private variable nFrames_

	private variable width_
	private variable height_
	private variable imageSize_
	private variable imageHeaderBin_
	private variable pane_

	constructor {} { setFPS 30 }

	method setFPS { fps } { set fps_ $fps }

	method fopen { fname } {
		set fd_ [open $fname "w"]
		fconfigure $fd_ -translation binary
		set fname_ $fname
		puts -nonewline $fd_ $headerTemplateBin_
		set nFrames_ 0

		set width_ 0
		set height_ 0
		set imageSize_ 0
		set imageHeaderBin_ ""
		return 1
	}

	method putFrame { imageName } {
		set thisWidth [image width $imageName]
		set thisHeight [image height $imageName]

		if { $nFrames_ == 0 } {
			set width_ $thisWidth
			set height_ $thisHeight
			set imageSize_ [expr {$width_ * $height_ * 4}]
			set imageHeaderBin_ [binary format "a4i" "00db" $imageSize_]
			set pane_ [image create photo -width $width_ -height $height_]
		} else {
			if { $width_ != $thisWidth } { return 0 }
			if { $height_ != $thisHeight } { return 0 }
		}				

		#have ppm binary image in variable "ppm"
		$pane_ copy $imageName -subsample 1 -1
		set ppm [::base64::decode [$pane_ data -format ppm]]
		scan $ppm "%s%s%s%s%n" dum_fmt dum_w dum_h dum_max headerLength

		#dump the binary to outbin
		set loc0 [expr $headerLength + 1]
		set loc1 [expr $headerLength + 2]
		set loc2 [expr $headerLength + 3]
		set locEnd [expr $loc0 + $width_ * $height_ * 3]
		while { $loc0 < $locEnd } {
			append outbin [string index $ppm $loc2] [string index $ppm $loc1] [string index $ppm $loc0] \0
			incr loc0 3
			incr loc1 3
			incr loc2 3
		}
		puts -nonewline $fd_ $imageHeaderBin_
		puts -nonewline $fd_ $outbin
		incr nFrames_
		return 1
	}

	method fclose {} {
		# generate index chunk
		puts -nonewline $fd_ [binary format "a4" "idx1"]
		puts -nonewline $fd_ [binary format "i" [expr {0x10 * $nFrames_}]]
		set loc 4
		for { set i 0 } { $i < $nFrames_ } { incr i } {
			puts -nonewline $fd_ [binary format "a4iii" "00db" 16 $loc $imageSize_]
			incr loc $imageSize_
			incr loc 8
		}

		# fill in the size fields
		set frameInterval [expr {int(1000000.0 / $fps_ )}]
		set maxDataRate [expr {int($imageSize_ * $fps_)}]
		set MoviListSize [expr {$nFrames_ * ($imageSize_ + 8) + 4}]
		set RIFFBodySize [expr {(0xe0 - 4 + $MoviListSize - 8) + (8 + $nFrames_ * 0x10) }]

		putInt32At $RIFFBodySize 0x04 
		putInt32At $frameInterval 0x20
		putInt32At $maxDataRate 0x24
		putInt32At $nFrames_ 0x30
		putInt32At $imageSize_ 0x3c
		putInt32At $width_ 0x40
		putInt32At $height_ 0x44
		putInt32At $frameInterval 0x80
		putInt32At $nFrames_ 0x8c
		putInt32At $imageSize_ 0x90
		putInt16At $width_ 0xa0
		putInt16At $height_ 0xa2
		putInt32At $width_ 0xb0
		putInt32At $height_ 0xb4
		putInt32At $imageSize_ 0xc0
		putInt32At $MoviListSize 0xd8
		close $fd_
		
		image delete $pane_
	}

	private method putInt32At { val offset } {
		seek $fd_ $offset start
		puts -nonewline $fd_ [binary format i $val]
	}

	private method putInt16At { val offset } {
		seek $fd_ $offset start
		puts -nonewline $fd_ [binary format s $val]
	}
}
